const fs         = require('fs');
const path       = require('path');
const glob       = require('glob');
const {execSync} = require('child_process');
const {parse}    = require('jsonc-parser');

const {commitChanges, pushChanges} = require("./upgrade-version");

function readProjects()
{
    try {
        const {projects} = parse(fs.readFileSync('angular.json', 'utf-8'));
        return Object.entries(projects)
            .filter(([, project]) => project.projectType === 'library')
            .map(([name, project]) =>
            {
                const packageJson = require(path.join(process.cwd(), project.root, 'package.json'));
                return {
                    name       : name,
                    packageName: packageJson.name,
                    root       : project.root,
                    version    : packageJson.version,
                };
            });
    } catch (error) {
        throw new Error(`Failed to read projects: ${error.message}`);
    }
}

function copyNpmrcToProjects(projects)
{
    try {
        projects.forEach((project) =>
        {
            const dest = path.join(project.root, '.npmrc');
            const npmrc = fs.readFileSync('.npmrc', 'utf-8');
            // copy to dest and uncomment the registry setting
            fs.writeFileSync(dest, npmrc.replace(/^;registry=/m, 'registry='));
        });
    } catch (error) {
        throw new Error(`Failed to copy .npmrc to projects: ${error.message}`);
    }
}

function runCommand(command, errorMessage, cwd = process.cwd())
{
    try {
        console.log(command);
        execSync(command, {stdio: 'inherit', cwd: cwd});
    } catch (error) {
        throw new Error(`${errorMessage}: ${error.message}`);
    }
}

function buildProjects(projects)
{
    projects.forEach((project) =>
    {
        runCommand(`ng build -c production ${project.name}`, `Failed to build project ${project.name}`);
    });
}

function publishProjects(projects)
{
    const tsConfig                    = fs.readFileSync('tsconfig.json', 'utf-8');
    const {compilerOptions: {outDir}} = parse(tsConfig);

    projects.forEach((project) =>
    {
        const distDir = path.join(project.root, outDir);
        const majorMinorPatch = project.version.match(/^[0-9]+\.[0-9]+\.[0-9]+/)?.[0];

        if (!majorMinorPatch) {
            throw new Error(`Invalid version string in ${project.name}`);
        }

        // Pack the project first
        runCommand(`npm pack`, `Failed to pack project ${project.name}`, distDir);

        // Retrieve package name assuming there is only one tarball in the directory
        const packedPackage = glob.sync(`${distDir}/*.tgz`)[0];

        if (!majorMinorPatch) {
            throw new Error(`Invalid version string in ${project.name}`);
        }

        runCommand(`npm publish --tag latest ${packedPackage}`, `Failed to publish project ${project.name}`);

        // Add version-major-minor-patch as an additional tag
        runCommand(`npm dist-tag add ${project.packageName}@${project.version} latest_${majorMinorPatch}`, `Failed to add dist-tag ${majorMinorPatch} to project ${project.name}.`);
    });
}

try {
    const projects = readProjects();
    copyNpmrcToProjects(projects);
    buildProjects(projects);
    publishProjects(projects);
    console.log('Successfully published projects:\n- ' + projects.map((project) => project.name + '@' + project.version).join('\n- '));

    // Get new version from main project's package.json
    const mainPackageJson = require(path.join(process.cwd(), 'package.json'));
    const newVersion      = mainPackageJson.version;
    commitChanges(newVersion);
    pushChanges();
    console.log(`Successfully committed and pushed version update with version ${newVersion}`);

} catch (error) {
    console.error(error);
    process.exit(1);
}
