const fs         = require('fs');
const glob       = require('glob');
const path       = require('path');
const {execSync} = require('child_process');
const semver     = require('semver');

const mainProjectPath = process.cwd();
const subProjectsPath = mainProjectPath //path.join(mainProjectPath, 'projects');
const upgradeType = process.argv[2];

const gitCommit  = 'publish(npm): new version';
const gitCommand = 'git -c credential.helper=';
const gitUsername = 'team@yukawa.de';

function getDirectories(source)
{
    return fs.readdirSync(source, {withFileTypes: true})
        .filter(dirent => dirent.isDirectory())
        .map(dirent => dirent.name)
        .filter(dir => fs.existsSync(path.join(source, dir, 'package.json')));
}

function upgradeVersion(version, upgradeType)
{
    const parsedVersion = semver.parse(version, {loose: true});

    switch (upgradeType) {
        case 'major':
            return `${parsedVersion.major + 1}.0.0`;
        case 'minor':
            return `${parsedVersion.major}.${parsedVersion.minor + 1}.0`;
        case 'patch':
            return `${parsedVersion.major}.${parsedVersion.minor}.${parsedVersion.patch + 1}`;
        case 'build':
            // use teamcity build number if available, otherwise increment build number
            const build = process.env.BUILD_NUMBER
                ? process.env.BUILD_NUMBER
                : parsedVersion.prerelease.length > 0
                    ? parsedVersion.prerelease[parsedVersion.prerelease.length - 1] + 1
                    : 0;
            return `${parsedVersion.major}.${parsedVersion.minor}.${parsedVersion.patch}-build.${build}`;
    }
}

function updatePackageJson(projectPath, upgradeType)
{
    const packageJsonPath = path.join(projectPath, 'package.json');
    try {
        const packageJson   = require(packageJsonPath);
        const oldVersion    = packageJson.version;
        const newVersion    = upgradeVersion(oldVersion, upgradeType);
        packageJson.version = newVersion;
        fs.writeFileSync(packageJsonPath, JSON.stringify(packageJson, null, 2) + '\n');
        console.log(`Upgraded ${path.basename(projectPath)} from ${oldVersion} to ${newVersion}`);
    } catch (error) {
        throw new Error(`An error occurred while updating ${projectPath}: ${error.message}`);
    }
}

function updatePackageLockJson(projectPath)
{
    try {
        if (fs.existsSync(path.join(projectPath, 'package-lock.json'))) {
            execSync('npm i --package-lock-only', {cwd: projectPath});
        } else {
            execSync('npm i --omit=optional', {cwd: projectPath});
        }
    } catch (error) {
        throw new Error(`An error occurred while updating package-lock.json in ${projectPath}: ${error.message}`);
    }
}

function commitChanges(newVersion)
{
    const packageFiles = glob.sync('**/package*.json', {ignore: ['node_modules/**', '**/target/angular/**']});
    const env          = Object.assign({}, process.env, {
        GIT_COMMITTER_NAME: gitUsername,
        GIT_COMMITTER_EMAIL: process.env.YUKAWA_GIT_USER,
        GIT_AUTHOR_NAME   : gitUsername,
        GIT_AUTHOR_EMAIL   : process.env.YUKAWA_GIT_USER
    });

    for (const file of packageFiles) {
        execSync(`git add ${file}`, {cwd: mainProjectPath});
    }
    execSync(`git commit -m "${gitCommit} ${newVersion}"`, {cwd: mainProjectPath, env});
}

function pushChanges()
{
    const gitAskPassFullPath = path.join(__dirname, 'git-askpass.sh');

    // Set the execution bit to make sure the script is executable
    fs.chmodSync(gitAskPassFullPath, '755');

    const env = Object.assign({}, process.env, {
        GIT_ASKPASS        : gitAskPassFullPath,
        GIT_TERMINAL_PROMPT: 0,  // Disable terminal prompts for credentials
    });

    // Get the name of the remote repository
    const remoteName = execSync('git remote').toString().split('\n')[0].trim();

    // Get the name of the current branch
    const currentBranch = execSync('git rev-parse --abbrev-ref HEAD', {cwd: mainProjectPath}).toString().trim();

    // Fetch the latest changes from the remote repository
    execSync(`${gitCommand} fetch ${remoteName} ${currentBranch}`, {cwd: mainProjectPath, env});

    // Check if rebase is needed
    const localCommit  = execSync(`git rev-parse ${currentBranch}`).toString().trim();
    const remoteCommit = execSync(`git rev-parse ${remoteName}/${currentBranch}`).toString().trim();

    if (localCommit !== remoteCommit) {
        // Find the root directory of the Git repository
        const gitRoot = execSync('git rev-parse --show-toplevel', {cwd: mainProjectPath}).toString().trim();

        // Check if a rebase is in progress and abort it if necessary
        const rebaseMergeDir = path.join(gitRoot, '.git', 'rebase-merge');
        if (fs.existsSync(rebaseMergeDir)) {
            execSync('git rebase --abort', {env});
        }

        // Rebase the current branch onto the corresponding remote branch
        execSync(`${gitCommand} rebase ${remoteName}/${currentBranch}`, {cwd: mainProjectPath, env});
    }

    // Push the changes, ensuring a fast-forward merge
    execSync(`${gitCommand} push`, {cwd: mainProjectPath, env});
}


function upgrade()
{
    try {
        // Upgrade main project
        updatePackageJson(mainProjectPath, upgradeType);
        updatePackageLockJson(mainProjectPath);

        // Upgrade subprojects
        const subProjects = getDirectories(subProjectsPath);
        subProjects.forEach(subProject =>
        {
            const subProjectPath = path.join(subProjectsPath, subProject);
            updatePackageJson(subProjectPath, upgradeType);
        });

        console.log('All projects have been successfully upgraded!');
    } catch (error) {
        console.error(error);
        process.exit(1);
    }
}


if (upgradeType) {
    // Upgrade all projects
    upgrade();
} else {
    // Commit and push changes after publishing manually
    module.exports = {
        commitChanges,
        pushChanges
    }
}
